/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAPBLOCK_HEADER
#define MAPBLOCK_HEADER

#include <jmutex.h>
#include <assert.h>
#include "common_irrlicht.h"
#include "mapnode.h"
#include "referencecounted.h"

//#define MAP_BLOCKSIZE 20
//#define MAP_BLOCKSIZE 32
#define MAP_BLOCKSIZE 16

// Named by looking towards z+
enum{
	FACE_BACK=0,
	FACE_TOP,
	FACE_RIGHT,
	FACE_FRONT,
	FACE_BOTTOM,
	FACE_LEFT
};

struct FastFace
{
	u8 material;
	/*u16 x;
	u16 y;
	u16 z;*/
	video::S3DVertex vertices[4]; // Precalculated vertices
	//u8 direction; // Index in the precalc buffer
};

class Map;

class MapBlock : public ReferenceCounted
{
	v3s16 m_pos; // Position in blocks
	//core::array<MapNode> data;
	MapNode * data;
	Map *map;
	bool changed;

public:

	core::list<FastFace*> *fastfaces;
	JMutex fastfaces_mutex;

	// This should be locked before doing anything with the block
	JMutex mutex;
	/*
		TODO: How about a BlockReference, or actually, a Reference<T>,
		which could handle a ReferenceCounted
	*/
	
	MapBlock(Map *a_map, v3s16 pos)
		: ReferenceCounted("MapBlock"), m_pos(pos), map(a_map), changed(true)
	{
		data = NULL;
		reallocate();
		//m_refcount_mutex.Init();
		mutex.Init();
		fastfaces = new core::list<FastFace*>;
		fastfaces_mutex.Init();
	}

	~MapBlock()
	{
		assert(getRefcount() == 0); // Destructing while referenced is not good.
		fastfaces_mutex.Lock();
		delete fastfaces;
		fastfaces_mutex.Unlock();
		if(data)
			free(data);
	}
	
	bool getChangedFlag()
	{
		return changed;
	}

	void resetChangedFlag()
	{
		/*v3s16 p = getPosRelative();
		std::cout<<"MapBlock("<<p.X<<","<<p.Y<<","<<p.Z<<")"
				<<"::resetChangedFlag()"
				<<std::endl;*/
				
		changed = false;
	}

	void setChangedFlag()
	{
		/*v3s16 p = getPosRelative();
		std::cout<<"MapBlock("<<p.X<<","<<p.Y<<","<<p.Z<<")"
				<<"::setChangedFlag()"
				<<std::endl;*/
				
		changed = true;
	}

	v3s16 getPosRelative()
	{
		//return posRelative;
		return m_pos * MAP_BLOCKSIZE;
	}
		
	//TODO: remove
	v3s16 getSizeNodes()
	{
		//return size;
		return v3s16(MAP_BLOCKSIZE, MAP_BLOCKSIZE, MAP_BLOCKSIZE);
	}

	core::aabbox3d<s16> getBox()
	{
		return core::aabbox3d<s16>(getPosRelative(),
				getPosRelative() + getSizeNodes() - v3s16(1,1,1));
	}
	
	void reallocate()
	{
		//data.clear();
		if(data != NULL)
			free(data);
		u32 l = MAP_BLOCKSIZE * MAP_BLOCKSIZE * MAP_BLOCKSIZE;
		//data.reallocate(l);
		data = (MapNode*)malloc(sizeof(MapNode) * l);
		for(u32 i=0; i<l; i++){
			//data.push_back(MapNode());
			data[i] = MapNode();
		}
	}

	MapNode * getNodePtr(s16 x, s16 y, s16 z)
	{
		if(x < 0 || x >= MAP_BLOCKSIZE) throw 1;
		if(y < 0 || y >= MAP_BLOCKSIZE) throw 1;
		if(z < 0 || z >= MAP_BLOCKSIZE) throw 1;
		return &data[z*MAP_BLOCKSIZE*MAP_BLOCKSIZE + y*MAP_BLOCKSIZE + x];
	}
	
	MapNode * getNodePtr(v3s16 p)
	{
		return getNodePtr(p.X, p.Y, p.Z);
	}
	
	MapNode & getNode(s16 x, s16 y, s16 z)
	{
		if(x < 0 || x >= MAP_BLOCKSIZE) throw 1;
		if(y < 0 || y >= MAP_BLOCKSIZE) throw 1;
		if(z < 0 || z >= MAP_BLOCKSIZE) throw 1;
		return data[z*MAP_BLOCKSIZE*MAP_BLOCKSIZE + y*MAP_BLOCKSIZE + x];
	}
	
	MapNode & getNode(v3s16 p)
	{
		return getNode(p.X, p.Y, p.Z);
	}
	
	void setNode(s16 x, s16 y, s16 z, MapNode & n)
	{
		if(x < 0 || x >= MAP_BLOCKSIZE) throw 1;
		if(y < 0 || y >= MAP_BLOCKSIZE) throw 1;
		if(z < 0 || z >= MAP_BLOCKSIZE) throw 1;
		data[z*MAP_BLOCKSIZE*MAP_BLOCKSIZE + y*MAP_BLOCKSIZE + x] = n;
	}
	
	void setNode(v3s16 p, MapNode & n)
	{
		setNode(p.X, p.Y, p.Z, n);
	}

	/*
		Uses the Map's function to get a node on any block
		with relative coordinates to this block.
	*/
	MapNode * getNodePtrMap(v3s16 p);

	/*
		Uses the Map's function to set a node on any block
		with relative coordinates to this block.
	*/
	void setNodeMap(v3s16 p, MapNode & n);

	void drawbox(s16 x0, s16 y0, s16 z0, s16 w, s16 h, s16 d, MapNode node)
	{
		for(u16 z=0; z<d; z++)
			for(u16 y=0; y<h; y++)
				for(u16 x=0; x<w; x++)
					setNode(x0+x, y0+y, z0+z, node);
	}

	static FastFace * makeFastFace(u8 material, light_t light, v3f p,
			v3f dir, v3f scale, v3f posRelative_f);
	
	light_t getFaceLight(v3s16 p, v3s16 face_dir);
	
	/*
		Gets node material from any place relative to block.
		Returns MATERIAL_AIR if doesn't exist.
	*/
	u8 getNodeMaterial(v3s16 p)
	{
		MapNode *n = getNodePtrMap(p);
		if(n == NULL)
			return MATERIAL_IGNORE;
		else
			return n->d;
	}

	/*
		startpos:
		translate_dir: unit vector with only one of x, y or z
		face_dir: unit vector with only one of x, y or z
	*/
	void updateFastFaceRow(v3s16 startpos,
			u16 length,
			v3s16 translate_dir,
			v3s16 face_dir,
			core::list<FastFace*> &dest);

	/*
		Propagates sunlight down through the block.

		If there is a block above, continues from it.

		If there is no block above, assumes there is sunlight.

		Returns true if some sunlight touches the bottom of the block.
	*/

	bool propagateSunlight()
	{
		bool sunlight_touches_bottom = false;

		for(s16 x=0; x<MAP_BLOCKSIZE; x++)
		{
			for(s16 z=0; z<MAP_BLOCKSIZE; z++)
			{
				// Check if node above block has sunlight
				MapNode *n = getNodePtrMap(v3s16(x, MAP_BLOCKSIZE, z));
				if(n != NULL){
					if(n->light < 0.999)
						// No sunlight here
						continue;
				}
				
				// Continue spreading sunlight downwards
				s16 y = MAP_BLOCKSIZE-1;
				for(; y >= 0; y--){
					v3s16 pos(x, y, z);
					
					MapNode *n = getNodePtr(pos);
					if(n == NULL)
						break;

					if(n->transparent())
					{
						n->light = 1.0;
					}
					else{
						break;
					}
				}
				if(y == -1)
					sunlight_touches_bottom = true;
			}
		}

		return sunlight_touches_bottom;
	}

	/*
		Find all faces being between material and air
	*/
	void updateFastFaces();

};

#endif

