/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef HEIGHTMAP_HEADER
#define HEIGHTMAP_HEADER

#include <iostream>
#include <assert.h>

#include "common_irrlicht.h"

#define GROUNDHEIGHT_NOTFOUND_SETVALUE (-10e6)
#define GROUNDHEIGHT_VALID_MINVALUE    ( -9e6)

#define HEIGHTMAP_DEBUGPRINT 0

class HeightmapSource
{
public:
	virtual f32 getGroundHeight(v2s16 p)
	{
		printf("HeightmapSource::getGroundHeight() stub called\n");
		assert(0);
	};
	virtual void setGroundHeight(v2s16 p, f32 h)
	{
		printf("HeightmapSource::setGroundHeight() stub called\n");
		assert(0);
	};
};

class HeightmapSingleSource : public HeightmapSource
{
	core::map<v2s16, f32> groundheights;
public:
	f32 getGroundHeight(v2s16 p)
	{
		//return 0.0;

		core::map<v2s16, f32>::Node *n = groundheights.find(p);
		if(n == NULL){
			/*printf("HSS::getGroundHeight((%i,%i)) = "
					"GROUNDHEIGHT_NOTFOUND_SETVALUE\n",
					p.X, p.Y);*/
			return GROUNDHEIGHT_NOTFOUND_SETVALUE;
		}
		else{
			/*printf("HSS::getGroundHeight((%i,%i)) = %f\n",
					p.X, p.Y, n->getValue());*/
			return n->getValue();
		}
	}

	void setGroundHeight(v2s16 p, f32 h)
	{
		/*printf("HSS::setGroundHeight((%i,%i), %f)\n",
				p.X, p.Y, h);*/
		if(h < GROUNDHEIGHT_VALID_MINVALUE){
			printf("HSS::setGroundHeight((%i,%i), %f): value set to invalid\n",
					p.X, p.Y, h);
		}
		assert(h > GROUNDHEIGHT_VALID_MINVALUE);
		groundheights[p] = h;
	}
};

class Heightmap
{
	s32 W;
	s32 H;
	HeightmapSource *m_source;
	s16 m_blocksize;
	v2s16 m_offset_blocks;

public:

	Heightmap(s32 w=0, s32 h=0)
	{
		W = 0;
		H = 0;
		m_source = NULL;
		m_blocksize = 0;
		m_offset_blocks = v2s16(0,0);
		
		if(w != 0 && h != 0)
			allocate(w, h);
	}

	~Heightmap()
	{
		unallocate();
	}

	void allocate(s32 w, s32 h)
	{
		unallocate();
		
		/*heightmap = (f32*)malloc(sizeof(f32)*w*h);
		if(heightmap == NULL) throw;
		//seeded = (u8*)malloc(sizeof(u8)*w*h);
		for(s32 i=0; i<w*h; i++){
			heightmap[i] = 0.0;
			//seeded[i] = false;
		}*/

		W = w;
		H = h;
	}

	void unallocate()
	{
		/*if(heightmap)
			free(heightmap);
		heightmap = NULL;*/
		/*if(seeded)
			free(seeded);
		seeded = NULL;*/
	}

	void setSource(HeightmapSource *source, s16 blocksize,
			v2s16 offset_blocks)
	{
		m_source = source;
		m_blocksize = blocksize;
		m_offset_blocks = offset_blocks;
	}

	/*
		TODO: BorderWrapper class or something to allow defining
		borders that wrap to an another heightmap. The algorithm
		should be allowed to edit stuff over the border and on
		the border in that case, too.
		This will allow non-square heightmaps, too. (probably)
	*/

	void print()
	{
		printf("Heightmap::print(): size is %ix%i\n", W, H);
		for(s32 y=0; y<H; y++){
			for(s32 x=0; x<W; x++){
				/*if(getSeeded(v2s16(x,y)))
					printf("S");*/
				f32 n = getNode(v2s16(x,y));
				if(n < GROUNDHEIGHT_VALID_MINVALUE)
					printf("  -  ");
				else
					printf("% -4.1f ", getNode(v2s16(x,y)));
			}
			printf("\n");
		}
	}
	
	bool overborder(v2s16 p)
	{
		return (p.X < 0 || p.X >= W || p.Y < 0 || p.Y >= H);
	}

	/*bool atborder(v2s16 p)
	{
		if(overborder(p))
			return false;
		return (p.X == 0 || p.X == W-1 || p.Y == 0 || p.Y == H-1);
	}*/

	f32 getNode(v2s16 p)
	{
		assert(m_source != NULL);
		// Place of block on source
		v2s16 blockpos_nodes = m_offset_blocks * m_blocksize;
		return m_source->getGroundHeight(blockpos_nodes + p);
	}

	f32 getGroundHeight(v2s16 p)
	{
		//TODO is this limit necessary?
		if(overborder(p)) return GROUNDHEIGHT_NOTFOUND_SETVALUE;
		return getNode(p);
	}
	
	void setNode(v2s16 p, f32 n)
	{
		assert(overborder(p) == false);
		assert(m_source != NULL);
		// Place of block on source
		v2s16 blockpos_nodes = m_offset_blocks * m_blocksize;
		m_source->setGroundHeight(blockpos_nodes + p, n);
	}

	void seedNode(v2s16 p, f32 n)
	{
		setNode(p, n);
	}

	f32 avgNeighbours(v2s16 p, s16 d)
	{
		//printf("avgNeighbours((%i,%i), %i): ", p.X, p.Y, d);
		v2s16 dirs[4] = {
			v2s16(1,0),
			v2s16(0,1),
			v2s16(-1,0),
			v2s16(0,-1)
		};
		f32 sum = 0.0;
		f32 count = 0.0;
		for(u16 i=0; i<4; i++){
			v2s16 p2 = p + dirs[i] * d;
			/*if(overborder(p2))
				//TODO: get value from somewhere else
				continue;*/
			f32 n = getNode(p2);
			if(n < GROUNDHEIGHT_VALID_MINVALUE)
				continue;
			sum += n;
			count += 1.0;
			//printf("(%i,%i)=%f ", p2.X, p2.Y, n);
		}
		//printf("\n");
		assert(count > 0.001);
		return sum / count;
	}
	
	f32 avgDiagNeighbours(v2s16 p, s16 d)
	{
		/*printf("avgDiagNeighbours((%i,%i), %i): ",
			p.X, p.Y, d);*/
		v2s16 dirs[4] = {
			v2s16(1,1),
			v2s16(-1,-1),
			v2s16(-1,1),
			v2s16(1,-1)
		};
		f32 sum = 0.0;
		f32 count = 0.0;
		for(u16 i=0; i<4; i++){
			v2s16 p2 = p + dirs[i] * d;
			/*if(overborder(p2))
				//TODO: get value from somewhere else
				continue;*/
			f32 n = getNode(p2);
			if(n < GROUNDHEIGHT_VALID_MINVALUE)
				continue;
			sum += n;
			count += 1.0;
			//printf("(%i,%i)=%f ", p2.X, p2.Y, n);
		}
		//printf("\n");
		assert(count > 0.001);
		return sum / count;
	}
	
	/*
		Adds a point to transform into a diamond pattern
		a = 2, 4, 8, 16, ...
	*/
	void makeDiamond(v2s16 center, s16 a, f32 randmax)
	{
		f32 n = avgDiagNeighbours(center, a/2);
		n += ((float)(rand()%2000000)/1e6-1.0)*randmax;
		setNode(center, n);
	}

	/*
		Adds points to transform into a diamond pattern
		a = 2, 4, 8, 16, ...
	*/
	void makeDiamonds(s16 a, f32 randmax)
	{
		s16 count_y = (H-1) / a;
		s16 count_x = (W-1) / a;
		for(s32 yi=0; yi<count_y; yi++){
			for(s32 xi=0; xi<count_x; xi++){
				v2s16 center(xi*a + a/2, yi*a + a/2);
				makeDiamond(center, a, randmax);
			}
		}
	}

	/*
		Adds a point to transform into a square pattern
		a = 2, 4, 8, 16, ...
	*/
	void makeSquare(v2s16 center, s16 a, f32 randmax)
	{
		f32 n = avgNeighbours(center, a/2);
		n += ((float)(rand()%2000000)/1e6-1.0)*randmax;
		setNode(center, n);
	}

	/*
		Adds points to transform into a square pattern
		a = 2, 4, 8, 16, ...
	*/
	void makeSquares(s16 a, f32 randmax)
	{
		/*
			This is a bit tricky; the points have to be kind of
			interlaced. (but not exactly)
			
			On even rows (0, 2, 4, ...) the nodes are horizontally
			at a*n+a/2 positions, vertically at a*n.

			On odd rows (1, 3, 5, ...) the nodes are horizontally
			at a*n positions, vertically at a*n+a/2.
		*/

		s16 count_y = (H-1) / a + 1;
		for(s32 yi=0; yi<=count_y; yi++){
			s16 count_x;
			// Even rows
			count_x = (W-1) / a;
			for(s32 xi=0; xi<count_x; xi++){
				v2s16 center(xi*a + a/2, yi*a);
				makeSquare(center, a, randmax);
			}
			if(yi >= count_y - 1)
				break;
			// odd rows
			count_x = (W-1) / a + 1;
			for(s32 xi=0; xi<count_x; xi++){
				v2s16 center(xi*a, yi*a + a/2);
				makeSquare(center, a, randmax);
			}
		}
	}

	void DiamondSquare(f32 randmax, f32 randfactor)
	{
		u16 a;
		if(W < H)
			a = W-1;
		else
			a = H-1;
		
		// Check that a is a power of two
		if((a & (a-1)) != 0)
			throw;
		
		while(a >= 2)
		{
			makeDiamonds(a, randmax);
			if(HEIGHTMAP_DEBUGPRINT){
				printf("MakeDiamonds a=%i result:\n", a);
				print();
			}
			makeSquares(a, randmax);
			if(HEIGHTMAP_DEBUGPRINT){
				printf("MakeSquares a=%i result:\n", a);
				print();
			}
			a /= 2;
			randmax *= randfactor;
		}
	}
	
	//TODO
	/*
		source: Connects borders with this
		blocksize: Size of heightmap
		offset_blocks: placement of block on master
		master: source for corner heights
	*/
	void GenerateContinued(HeightmapSource *source, s16 blocksize,
			v2s16 offset_blocks, Heightmap *master, f32 default_value,
			f32 randmax, f32 randfactor);
};

#endif

