#include "server.h"
#include "utility.h"
#include <iostream>
#include "clientserver.h"
#include "map.h"
#include "jmutexautolock.h"
#include "main.h"
#include "constants.h"

void * ServerThread::Thread()
{
	ThreadStarted();

	while(getRun())
	{
		try{
			m_server->AsyncRunStep();
		
			//dout_server<<"Running m_server->Receive()"<<std::endl;
			m_server->Receive();
		}
		catch(con::NoIncomingDataException &e)
		{
		}
		catch(std::exception &e)
		{
			/*dout_server<<"ServerThread: Some exception: "
					<<e.what()<<std::endl;*/
			derr_server<<"ServerThread: Some exception: "
					<<e.what()<<std::endl;
		}
	}
	

	return NULL;
}

void * EmergeThread::Thread()
{
	ThreadStarted();

	/*
		TODO
		Get block info from queue, emerge them and send them
		to clients.

		After queue is empty, exit.
	*/
	while(getRun())
	{
		QueuedBlockEmerge *qptr = m_server->m_emerge_queue.pop();
		if(qptr == NULL)
			break;
		
		SharedPtr<QueuedBlockEmerge> q(qptr);

		v3s16 &p = q->pos;
		
		//derr_server<<"EmergeThread::Thread(): running"<<std::endl;
		
		/*
			Try to emerge it from somewhere.

			If it is only wanted as optional, only loading from disk
			will be allowed.
		*/
		
		/*
			Check if any peer wants it as non-optional. In that case it
			will be generated.
		*/
		bool optional = true;
		{
			core::map<u16, u8>::Iterator i;
			for(i=q->peer_ids.getIterator(); i.atEnd()==false; i++)
			{
				u8 flags = i.getNode()->getValue();
				if((flags & TOSERVER_GETBLOCK_FLAG_OPTIONAL) == false)
					optional = false;
			}
		}
		
		MapBlock *block = NULL;
		try{
			bool only_from_disk = false;
			
			if(optional)
				only_from_disk = true;
			
			//TODO: This mutex blocks almost everything
			//NOTE: This might work nicely without this because this
			//      is the only place where we are calling emergeBlock
			//JMutexAutoLock lock(m_server->m_env_mutex);

			ServerMap &map = ((ServerMap&)m_server->m_env.getMap());
			block = map.emergeBlock(p, only_from_disk);
		}
		catch(InvalidPositionException &e)
		{
			// Block not found.
			// This happens when getting it is optional.
			assert(optional == true);
		}

		/*
			Send block to all listed peers
		*/
		
		JMutexAutoLock lock(m_server->m_con_mutex);

		if(optional || block == NULL)
		{
			SharedBuffer<u8> reply(8);
			writeU16(&reply[0], TOCLIENT_OPT_BLOCK_NOT_FOUND);
			writeS16(&reply[2], p.X);
			writeS16(&reply[4], p.Y);
			writeS16(&reply[6], p.Z);

			core::map<u16, u8>::Iterator i;
			for(i=q->peer_ids.getIterator(); i.atEnd()==false; i++)
			{
				u16 peer_id = i.getNode()->getKey();
				// Send as reliable
				m_server->m_con.Send(peer_id, 1, reply, true);
			}
		}
		else
		{
			core::map<u16, u8>::Iterator i;
			for(i=q->peer_ids.getIterator(); i.atEnd()==false; i++)
			{
				u16 peer_id = i.getNode()->getKey();

				/*
					Get peer serialization version
				*/
				con::Peer *peer;
				try{
					peer = m_server->m_con.GetPeer(peer_id);
				}
				catch(con::PeerNotFoundException &e)
				{
					derr_server<<DTIME<<"EmergeThread: Cancelling sending: "
							"peer "<<peer_id<<" not found"<<std::endl;
					continue;
				}
				
				u8 peer_ser_ver = peer->serialization_version;

				/*
					Create a packet with the block in the right format
				*/
				
				SharedBuffer<u8> blockdata = block->serialize(peer_ser_ver);
				u32 replysize = 8 + blockdata.getSize();
				SharedBuffer<u8> reply(replysize);
				writeU16(&reply[0], TOCLIENT_BLOCKDATA);
				writeS16(&reply[2], p.X);
				writeS16(&reply[4], p.Y);
				writeS16(&reply[6], p.Z);
				memcpy(&reply[8], *blockdata, blockdata.getSize());
				
				/*
					Send packet
				*/
				m_server->m_con.Send(peer_id, 1, reply, true);
			}
		}
	}

	return NULL;
}

PlayerInfo::PlayerInfo()
{
	name[0] = 0;
}

void PlayerInfo::PrintLine(std::ostream *s)
{
	(*s)<<id<<": \""<<name<<"\" ("
			<<position.X<<","<<position.Y
			<<","<<position.Z<<") ";
	address.print(s);
	(*s)<<" avg_rtt="<<avg_rtt;
	(*s)<<std::endl;
}

u32 PIChecksum(core::list<PlayerInfo> &l)
{
	core::list<PlayerInfo>::Iterator i;
	u32 checksum = 1;
	u32 a = 10;
	for(i=l.begin(); i!=l.end(); i++)
	{
		checksum += a * (i->id+1);
		checksum ^= 0x435aafcd;
		a *= 10;
	}
	return checksum;
}

Server::Server(std::string mapsavedir):
	m_env(new ServerMap(mapsavedir), dout_server),
	m_con(PROTOCOL_ID, 512, 20.0),
	m_thread(this),
	m_emergethread(this)
{
	m_env_mutex.Init();
	m_con_mutex.Init();
	m_step_dtime_mutex.Init();
	m_step_dtime = 0.0;
}

Server::~Server()
{
	stop();
}

void Server::start(unsigned short port)
{
	// Stop thread if already running
	m_thread.stop();
	
	// Initialize connection
	m_con.setTimeoutMs(50);
	m_con.Serve(port);

	// Start thread
	m_thread.setRun(true);
	m_thread.Start();
	
	dout_server<<"Server started on port "<<port<<std::endl;
}

void Server::stop()
{
	// Stop threads (set run=false first so both start stopping)
	m_thread.setRun(false);
	m_emergethread.setRun(false);
	m_thread.stop();
	m_emergethread.stop();
	
	dout_server<<"Server stopped"<<std::endl;
}

void Server::step(float dtime)
{
	{
		JMutexAutoLock lock(m_step_dtime_mutex);
		m_step_dtime += dtime;
	}
}

void Server::AsyncRunStep()
{
	float dtime;
	{
		JMutexAutoLock lock1(m_step_dtime_mutex);
		dtime = m_step_dtime;
		m_step_dtime = 0.0;
	}
	{
		// Process connection's timeouts
		JMutexAutoLock lock2(m_con_mutex);
		m_con.RunTimeouts(dtime);
	}
	{
		// Step environment
		// This also runs Map's timers
		JMutexAutoLock lock(m_env_mutex);
		m_env.step(dtime);
	}
	
	/*
		Do background stuff
	*/
	
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 5.0)
		{
			counter = 0.0;

			RemoveDisconnectedPlayers();
		}
	}
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 0.2)
		{
			counter = 0.0;

			SendPlayerPositions();
		}
	}
	{
		static float counter = 0.0;
		counter += dtime;
		if(counter >= 10.0)
		{
			counter = 0.0;

			SendPlayerInfos();
		}
	}

	{
		// Save map
		static float counter = 0.0;
		counter += dtime;
		if(counter >= SERVER_MAP_SAVE_INTERVAL)
		{
			counter = 0.0;

			JMutexAutoLock lock(m_env_mutex);
			// Save only changed parts
			m_env.getMap().save(true);
		}
	}
}

void Server::Receive()
{
	u32 data_maxsize = 10000;
	Buffer<u8> data(data_maxsize);
	u16 peer_id;
	u32 datasize;
	try{
		{
			JMutexAutoLock lock(m_con_mutex);
			datasize = m_con.Receive(peer_id, *data, data_maxsize);
		}
		ProcessData(*data, datasize, peer_id);
	}
	catch(con::InvalidIncomingDataException &e)
	{
		derr_server<<"Server::Receive(): "
				"InvalidIncomingDataException: what()="
				<<e.what()<<std::endl;
	}
	catch(con::PeerNotFoundException &e)
	{
		// The peer has been disconnected.
		// Find the associated player and remove it.

		JMutexAutoLock envlock(m_env_mutex);

		dout_server<<"ServerThread: peer_id="<<peer_id
				<<" has apparently closed connection. "
				<<"Removing player."<<std::endl;

		m_env.removePlayer(peer_id);
	}
}

void Server::ProcessData(u8 *data, u32 datasize, u16 peer_id)
{
	// Let free access to the environment and the connection.
	// Blocking them doesn't matter too much because this is the
	// server and latency is not such a problem.
	JMutexAutoLock envlock(m_env_mutex);
	JMutexAutoLock conlock(m_con_mutex);
	
	con::Peer *peer;
	try{
		peer = m_con.GetPeer(peer_id);
	}
	catch(con::PeerNotFoundException &e)
	{
		derr_server<<DTIME<<"Server::ProcessData(): Cancelling: peer "
				<<peer_id<<" not found"<<std::endl;
		return;
	}
	
	u8 peer_ser_ver = peer->serialization_version;

	try
	{

	if(datasize < 2)
		return;

	ToServerCommand command = (ToServerCommand)readU16(&data[0]);
	
	if(command == TOSERVER_INIT)
	{
		// [0] u16 TOSERVER_INIT
		// [2] u8 SER_FMT_VER_HIGHEST
		// [3] u8[20] player_name

		if(datasize < 3)
			return;

		// First byte after command is maximum supported
		// serialization version
		u8 client_max = data[2];
		u8 our_max = SER_FMT_VER_HIGHEST;
		// Use the highest version supported by both
		u8 deployed = core::min_(client_max, our_max);
		// If it's lower than the lowest supported, give up.
		if(deployed < SER_FMT_VER_LOWEST)
			deployed = SER_FMT_VER_INVALID;

		peer->serialization_version = deployed;

		/*
			Create a player
		*/

		Player *player = m_env.getPlayer(peer_id);

		// Create a player if it doesn't exist
		if(player == NULL)
		{
			dout_server<<"Server::ProcessData(): Adding player "
					<<peer_id<<std::endl;
			player = new Player(false);
			player->peer_id = peer_id;
			// update name if it was supplied
			if(datasize >= 20+3)
			{
				data[20+3-1] = 0;
				player->updateName((const char*)&data[3]);
			}
			m_env.addPlayer(player);
		}

		// Now answer with a TOCLIENT_INIT
		
		SharedBuffer<u8> reply(3);
		writeU16(&reply[0], TOCLIENT_INIT);
		writeU8(&reply[2], deployed);
		// Send as reliable
		m_con.Send(peer_id, 0, reply, true);

		return;
	}


	if(peer_ser_ver == SER_FMT_VER_INVALID)
	{
		derr_server<<DTIME<<"Server::ProcessData(): Cancelling: Peer"
				" serialization format invalid or not initialized."
				" Skipping incoming command="<<command<<std::endl;
		return;
	}
	
	Player *player = m_env.getPlayer(peer_id);

	if(player == NULL){
		derr_server<<"Server::ProcessData(): Cancelling: "
				"No player for peer_id="<<peer_id
				<<std::endl;
		return;
	}

	if(command == TOSERVER_GETBLOCK)
	{
		// Check for too short data
		if(datasize < 8)
			return;
		/*
			The client is asking a MapBlock.
			Get block data and send it.
			If not found, generate if not optional.
		*/

		/*
			[0] u16 command
			[2] u16 X
			[4] u16 Y
			[6] u16 Z
			[8] u8 flags (old versions leave this out)
		*/

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);
		
		u8 flags = 0;
		if(datasize >= 9)
			flags = data[8];

		/*dout_server<<"Server::ProcessData(): GETBLOCK"
				<<std::endl;*/
		/*std::cout<<"Server::ProcessData(): GETBLOCK"
				<<"("<<p.X<<","<<p.Y<<","<<p.Z<<")"<<std::endl;*/

		MapBlock *block;
		try{
			block = m_env.getMap().getBlockNoCreate(p);
		}
		catch(InvalidPositionException &e)
		{
			// Block not found.
			// Add it to the emerge queue, trigger the thread and quit.
			m_emerge_queue.addBlock(peer_id, p, flags);

			/*derr_server<<"Queued block "
					"("<<p.X<<","<<p.Y<<","<<p.Z<<")."
					" triggering thread."
					<<std::endl;*/

			m_emergethread.trigger();
			return;
		}

		SharedBuffer<u8> blockdata = block->serialize(peer_ser_ver);

		/*std::cout<<"Block serialized: size="<<blockdata.getSize()<<std::endl;
		for(u32 i=0; i<blockdata.getSize(); i++)
		{
			printf("%x ", (unsigned int)blockdata[i] & 0xff);
		}
		std::cout<<std::endl;*/

		u32 replysize = 8 + blockdata.getSize();

		//std::cout<<"replysize="<<replysize<<std::endl;

		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_BLOCKDATA);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		memcpy(&reply[8], *blockdata, blockdata.getSize());

		// Send as reliable
		m_con.Send(peer_id, 1, reply, true);
	}
	else if(command == TOSERVER_REMOVENODE)
	{
		if(datasize < 8)
			return;
			
		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		/*MapNode n;
		n.d = MATERIAL_AIR;
		try{
			m_env.getMap().setNode(p, n);*/
		try{
			m_env.getMap().removeNodeAndUpdate(p);
		}
		catch(InvalidPositionException &e)
		{
			derr_server<<"Server: Ignoring REMOVENODE: Node not found"
					<<std::endl;
			return;
		}

		u32 replysize = 8;
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_REMOVENODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
	}
	else if(command == TOSERVER_ADDNODE)
	{
		if(datasize < 8 + MapNode::serializedLength(peer_ser_ver))
			return;

		v3s16 p;
		p.X = readS16(&data[2]);
		p.Y = readS16(&data[4]);
		p.Z = readS16(&data[6]);

		/*MapNode n;
		n.deSerialize(&data[8], peer_ser_ver);
		try{
			m_env.getMap().setNode(p, n);*/
		MapNode n;
		n.deSerialize(&data[8], peer_ser_ver);
		try{
			u8 light = m_env.getMap().getNode(p).getLight();
			m_env.getMap().setNode(p, n);
			m_env.getMap().nodeAddedUpdate(p, light);
		}
		catch(InvalidPositionException &e)
		{
			derr_server<<"Server: Ignoring ADDNODE: Node not found"
					<<std::endl;
			return;
		}

		u32 replysize = 8 + MapNode::serializedLength(peer_ser_ver);
		SharedBuffer<u8> reply(replysize);
		writeU16(&reply[0], TOCLIENT_ADDNODE);
		writeS16(&reply[2], p.X);
		writeS16(&reply[4], p.Y);
		writeS16(&reply[6], p.Z);
		n.serialize(&reply[8], peer_ser_ver);
		// Send as reliable
		m_con.SendToAll(0, reply, true);
	}
	else if(command == TOSERVER_PLAYERPOS)
	{
		if(datasize < 2+12+12)
			return;

		//player->timeout_counter = 0.0;
		
		v3s32 ps = readV3S32(&data[2]);
		v3s32 ss = readV3S32(&data[2+12]);
		v3f position((f32)ps.X/100., (f32)ps.Y/100., (f32)ps.Z/100.);
		v3f speed((f32)ss.X/100., (f32)ss.Y/100., (f32)ss.Z/100.);
		player->setPosition(position);
		player->speed = speed;

		/*dout_server<<"Server::ProcessData(): Moved player "<<peer_id
				<<" to ("<<position.X
				<<","<<position.Y
				<<","<<position.Z
				<<")"<<std::endl;*/
	}
	/*else if(command == TOSERVER_PLAYERINFO)
	{
		if(datasize < 2+PLAYERNAME_SIZE)
			return;

		player->updateName((char*)&data[2]);
	}*/
	else
	{
		derr_server<<"WARNING: Server::ProcessData(): Ingoring "
				"unknown command "<<command<<std::endl;
	}
	
	} //try
	catch(SendFailedException &e)
	{
		derr_server<<"Server::ProcessData(): SendFailedException: "
				<<"what="<<e.what()
				<<std::endl;
	}
}

void Server::RemoveDisconnectedPlayers()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	core::list<Player*> players = m_env.getPlayers();
	
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		Player *player = *i;
		u16 peer_id = player->peer_id;
		con::Peer *peer = m_con.GetPeerNoEx(peer_id);
		if(peer != NULL)
			continue;
		
		dout_server<<"Server::RemoveDisconnectedPlayers(): "
				<<"peer_id="<<peer_id
				<<" has apparently closed connection. "
				<<"Removing associated player."<<std::endl;
		m_env.removePlayer(peer_id);
	}
}

void Server::SendPlayerPositions()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	core::list<Player*> players = m_env.getPlayers();
	
	u32 player_count = players.getSize();
	u32 datasize = 2+(2+12+12)*player_count;

	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOCLIENT_PLAYERPOS);
	
	u32 start = 2;
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		Player *player = *i;
		
		v3f pf = player->getPosition();
		v3s32 position(pf.X*100, pf.Y*100, pf.Z*100);
		v3f sf = player->speed;
		v3s32 speed(sf.X*100, sf.Y*100, sf.Z*100);

		writeU16(&data[start], player->peer_id);
		writeV3S32(&data[start+2], position);
		writeV3S32(&data[start+2+12], speed);
		start += 2+12+12;
	}

	JMutexAutoLock conlock(m_con_mutex);

	// Send as unreliable
	m_con.SendToAll(0, data, false);
}

void Server::SendPlayerInfos()
{
	JMutexAutoLock envlock(m_env_mutex);
	
	core::list<Player*> players = m_env.getPlayers();
	
	u32 player_count = players.getSize();
	u32 datasize = 2+(2+PLAYERNAME_SIZE)*player_count;

	SharedBuffer<u8> data(datasize);
	writeU16(&data[0], TOCLIENT_PLAYERINFO);
	
	u32 start = 2;
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		Player *player = *i;
		
		writeU16(&data[start], player->peer_id);
		snprintf((char*)&data[start+2], PLAYERNAME_SIZE, "%s", player->getName());
		start += 2+PLAYERNAME_SIZE;
	}

	JMutexAutoLock conlock(m_con_mutex);

	// Send as reliable
	m_con.SendToAll(0, data, true);
}

core::list<PlayerInfo> Server::getPlayerInfo()
{
	JMutexAutoLock envlock(m_env_mutex);
	JMutexAutoLock conlock(m_con_mutex);
	
	core::list<PlayerInfo> list;

	core::list<Player*> players = m_env.getPlayers();
	
	core::list<Player*>::Iterator i;
	for(i = players.begin();
			i != players.end(); i++)
	{
		PlayerInfo info;

		Player *player = *i;
		try{
			con::Peer *peer = m_con.GetPeer(player->peer_id);
			info.id = peer->id;
			info.address = peer->address;
			info.avg_rtt = peer->avg_rtt;
		}
		catch(con::PeerNotFoundException &e)
		{
			// Outdated peer info
			info.id = 0;
			info.address = Address(0,0,0,0,0);
			info.avg_rtt = 0.0;
		}

		snprintf(info.name, PLAYERNAME_SIZE, "%s", player->getName());
		info.position = player->getPosition();

		list.push_back(info);
	}

	return list;
}


