#ifndef UTILITY_HEADER
#define UTILITY_HEADER

// This is actually only included for u8, u16, etc.
#include "common_irrlicht.h"
#include <iostream>
#include <assert.h>

extern const v3s16 g_26dirs[26];

//#define min(x,y) (((x)<(y))?(x):(y))

void writeU32(u8 *data, u32 i);
void writeU16(u8 *data, u16 i);
void writeU8(u8 *data, u8 i);
u32 readU32(u8 *data);
u16 readU16(u8 *data);
u8 readU8(u8 *data);

void writeV3S32(u8 *data, v3s32 p);
v3s32 readV3S32(u8 *data);

void writeV2S16(u8 *data, v2s16 p);
v2s16 readV2S16(u8 *data);

// Inlines for signed variants of the above

inline void writeS32(u8 *data, s32 i){
	writeU32(data, (u32)i);
}
inline s32 readS32(u8 *data){
	return (s32)readU32(data);
}

inline void writeS16(u8 *data, s16 i){
	writeU16(data, (u16)i);
}
inline s16 readS16(u8 *data){
	return (s16)readU16(data);
}

/*
	None of these are used at the moment
*/

template <typename T>
class SharedPtr
{
public:
	SharedPtr(T *t=NULL)
	{
		refcount = new int;
		*refcount = 1;
		ptr = t;
	}
	SharedPtr(SharedPtr<T> &t)
	{
		*this = t;
	}
	~SharedPtr()
	{
		drop();
	}
	SharedPtr<T> & operator=(T *t)
	{
		drop();
		refcount = new int;
		*refcount = 1;
		ptr = t;
		return *this;
	}
	SharedPtr<T> & operator=(SharedPtr<T> &t)
	{
		drop();
		refcount = t.refcount;
		(*refcount)++;
		ptr = t.ptr;
		return *this;
	}
	T* operator->()
	{
		return ptr;
	}
	T & operator*()
	{
		return *ptr;
	}
	bool operator!=(T *t)
	{
		return ptr != t;
	}
	bool operator==(T *t)
	{
		return ptr == t;
	}
private:
	void drop()
	{
		assert((*refcount) > 0);
		(*refcount)--;
		if(*refcount == 0)
		{
			delete refcount;
			if(ptr != NULL)
				delete ptr;
		}
	}
	T *ptr;
	int *refcount;
};

template <typename T>
class Buffer
{
public:
	Buffer(unsigned int size)
	{
		m_size = size;
		data = new T[size];
	}
	Buffer(const Buffer &buffer)
	{
		m_size = buffer.m_size;
		data = new T[buffer.m_size];
		memcpy(data, buffer.data, buffer.m_size);
	}
	Buffer(T *t, unsigned int size)
	{
		m_size = size;
		data = new T[size];
		memcpy(data, t, size);
	}
	~Buffer()
	{
		delete[] data;
	}
	T & operator[](unsigned int i) const
	{
		return data[i];
	}
	T * operator*() const
	{
		return data;
	}
	unsigned int getSize() const
	{
		return m_size;
	}
private:
	T *data;
	unsigned int m_size;
};

template <typename T>
class SharedBuffer
{
public:
	SharedBuffer(unsigned int size)
	{
		m_size = size;
		data = new T[size];
		refcount = new unsigned int;
		(*refcount) = 1;
	}
	SharedBuffer(const SharedBuffer &buffer)
	{
		//std::cout<<"SharedBuffer(const SharedBuffer &buffer)"<<std::endl;
		m_size = buffer.m_size;
		data = buffer.data;
		refcount = buffer.refcount;
		(*refcount)++;
	}
	SharedBuffer & operator=(const SharedBuffer & buffer)
	{
		//std::cout<<"SharedBuffer & operator=(const SharedBuffer & buffer)"<<std::endl;
		if(this == &buffer)
			return *this;
		drop();
		m_size = buffer.m_size;
		data = buffer.data;
		refcount = buffer.refcount;
		(*refcount)++;
		return *this;
	}
	/*
		Copies whole buffer
	*/
	SharedBuffer(T *t, unsigned int size)
	{
		m_size = size;
		data = new T[size];
		memcpy(data, t, size);
		refcount = new unsigned int;
		(*refcount) = 1;
	}
	/*
		Copies whole buffer
	*/
	SharedBuffer(const Buffer<T> &buffer)
	{
		m_size = buffer.m_size;
		data = new T[buffer.getSize()];
		memcpy(data, *buffer, buffer.getSize());
		refcount = new unsigned int;
		(*refcount) = 1;
	}
	~SharedBuffer()
	{
		drop();
	}
	T & operator[](unsigned int i) const
	{
		return data[i];
	}
	T * operator*() const
	{
		return data;
	}
	unsigned int getSize() const
	{
		return m_size;
	}
private:
	void drop()
	{
		assert((*refcount) > 0);
		(*refcount)--;
		if(*refcount == 0)
		{
			delete[] data;
			delete refcount;
		}
	}
	T *data;
	unsigned int m_size;
	unsigned int *refcount;
};

inline SharedBuffer<u8> SharedBufferFromString(const char *string)
{
	SharedBuffer<u8> b((u8*)string, strlen(string)+1);
	return b;
}

/*
	TimeTaker
*/

class TimeTaker
{
public:
	TimeTaker(const char *name, IrrlichtDevice *dev)
	{
		m_name = name;
		m_dev = dev;
		m_time1 = m_dev->getTimer()->getRealTime();
	}
	~TimeTaker()
	{
		u32 time2 = m_dev->getTimer()->getRealTime();
		u32 dtime = time2 - m_time1;
		std::cout<<m_name<<" took "<<dtime<<"ms"<<std::endl;
	}
private:
	const char *m_name;
	IrrlichtDevice *m_dev;
	u32 m_time1;
};

// Calculates the borders of a "d-radius" cube
inline void getFacePositions(core::list<v3s16> &list, u16 d)
{
	if(d == 0)
	{
		list.push_back(v3s16(0,0,0));
		return;
	}

	// Take the top and bottom face without any borders
	// -d+1<x<d-1, y=+-d, -d+1<z<d-1
	for(s16 x=-d+1; x<=d-1; x++)
	for(s16 z=-d+1; z<=d-1; z++)
	{
		list.push_back(v3s16(x,d,z));
		list.push_back(v3s16(x,-d,z));
	}
	
	// Take the back and front face
	// -d<x<d, -d<y<d, z=+-d
	for(s16 x=-d; x<=d; x++)
	for(s16 y=-d; y<=d; y++)
	{
		list.push_back(v3s16(x,y,d));
		list.push_back(v3s16(x,y,-d));
	}

	// Take the left and right face without the already added blocks
	// x=+-d, -d<y<d, -d+1<z<d-1
	for(s16 y=-d; y<=d; y++)
	for(s16 z=-d+1; z<=d-1; z++)
	{
		list.push_back(v3s16(d,y,z));
		list.push_back(v3s16(-d,y,z));
	}

}

#endif

