/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAPNODE_HEADER
#define MAPNODE_HEADER

#include <iostream>
#include "common_irrlicht.h"
#include "light.h"
#include "utility.h"
#include "exceptions.h"
#include "serialization.h"

// Size of node in rendering units
#define BS 10

#define MATERIALS_COUNT 254

// This is completely ignored. It doesn't create faces with anything.
#define MATERIAL_IGNORE 255
// This is the common material through which the player can walk
// and which is transparent to light
#define MATERIAL_AIR 254

#define USEFUL_MATERIAL_COUNT 6

/*
	Materials-todo:

	GRAVEL
	  - Dynamics of gravel: if there is a drop of more than two
	    blocks on any side, it will drop in there. Is this doable?
*/

enum Material
{
	MATERIAL_STONE=0,

	MATERIAL_GRASS,

	/*
		For water, the param is water pressure. 0...127.
		
		- Water will be a bit like light, but with different flow
		  behavior.
		- Water blocks will fall down if there is empty space below.
		- If there is water below, the pressure of the block below is
		  the pressure of the current block + 1, or higher.
		- If there is any pressure in a horizontally neighboring
		  block, a water block will try to move away from it.
		- If there is >=2 of pressure in a block below, water will
		  try to move upwards.
		- NOTE: To keep large operations fast, we have to keep a
		        cache of the water-air-surfaces, just like with light
	*/
	MATERIAL_WATER,

	/*
		TORCH is transparent() altough you can't see through it
	*/
	MATERIAL_TORCH,

	MATERIAL_TREE,
	MATERIAL_LEAVES
};

struct MapNode
{
	//TODO: block type to differ from material
	//      (e.g. grass edges or something)
	// block type
	u8 d;

	// Removed because light is now stored in param for air
	// f32 light;

	/*
		Misc parameter. Initialized to 0.
		- For transparent() blocks, this is light intensity,
		  stored logarithmically from 0 to LIGHT_MAX.
		  Sunlight is LIGHT_SUN, which is LIGHT_MAX+1.
	*/
	s8 param;

	MapNode(const MapNode & n)
	{
		*this = n;
	}
	
	/*MapNode(u8 data=MATERIAL_AIR, light_t a_light=NODE_INITIAL_LIGHT, u8 a_param=0)
	{
		d = data;
		light = a_light;
		param = a_param;
	}*/

	MapNode(u8 data=MATERIAL_AIR, u8 a_param=0)
	{
		d = data;
		param = a_param;
	}

	/*
		If true, the node allows light propagation
	*/
	bool transparent()
	{
		/*
			TODO: Test changing torch to non-transparent. Then fix bugs for
			      that case.
		*/
		return (d == MATERIAL_AIR || d == MATERIAL_TORCH);
	}

	u8 light_source()
	{
		/*
			Note that a block that isn't transparent() can be a light source.
		*/
		if(d == MATERIAL_TORCH)
			return LIGHT_MAX;
		
		return 0;
	}

	u8 getLight()
	{
		// Select the brightest of [light_source, transparent_light]
		u8 light = 0;
		if(transparent())
			light = param;
		if(light_source() > param)
			light = light_source();
		return light;
	}

	void setLight(u8 a_light)
	{
		// If not transparent, can't set light
		if(transparent() == false)
			return;
		param = a_light;
	}

	static u32 serializedLength(u8 version)
	{
		if(!ser_ver_supported(version))
			throw VersionMismatchException("ERROR: MapNode format not supported");
			
		if(version == 0)
			return 1;
		else
			return 2;
	}
	void serialize(u8 *dest, u8 version)
	{
		if(!ser_ver_supported(version))
			throw VersionMismatchException("ERROR: MapNode format not supported");
			
		if(version == 0)
		{
			dest[0] = d;
		}
		else
		{
			dest[0] = d;
			dest[1] = param;
		}
	}
	void deSerialize(u8 *source, u8 version)
	{
		if(!ser_ver_supported(version))
			throw VersionMismatchException("ERROR: MapNode format not supported");
			
		if(version == 0)
		{
			d = source[0];
		}
		else if(version == 1)
		{
			d = source[0];
			// This version doesn't support saved lighting
			if(transparent() || light_source() > 0)
				param = 0;
			else
				param = source[1];
		}
		else
		{
			d = source[0];
			param = source[1];
		}
	}
};



#endif

