#include "mapsector.h"
#include "jmutexautolock.h"
#include "client.h"
#include "exceptions.h"

MapSector::MapSector(NodeContainer *parent, v2s16 pos):
		m_parent(parent),
		m_pos(pos),
		m_block_cache(NULL),
		m_heightmap(NULL),
		m_objects(NULL),
		//m_highest_y(-32768),
		differs_from_disk(true),
		usage_timer(0.0)
{
	m_mutex.Init();
	assert(m_mutex.IsInitialized());
}

MapSector::~MapSector()
{
	//std::cout<<"MapSector destructor called"<<std::endl;
	core::map<s16, MapBlock*>::Iterator i = m_blocks.getIterator();
	for(; i.atEnd() == false; i++)
	{
		delete i.getNode()->getValue();
	}

	if(m_heightmap)
		delete m_heightmap;

	if(m_objects)
		delete m_objects;
}

MapBlock * MapSector::getBlockBuffered(s16 y)
{
	MapBlock *block;

	if(m_block_cache != NULL && y == m_block_cache_y){
		return m_block_cache;
	}
	
	// If block doesn't exist, return NULL
	core::map<s16, MapBlock*>::Node *n = m_blocks.find(y);
	if(n == NULL)
	{
		block = NULL;
	}
	// If block exists, return it
	else{
		block = n->getValue();
	}
	
	// Cache the last result
	m_block_cache_y = y;
	m_block_cache = block;
	
	return block;
}

MapBlock * MapSector::getBlockNoCreate(s16 y)
{
	JMutexAutoLock lock(m_mutex);
	
	MapBlock *block = getBlockBuffered(y);

	if(block == NULL)
		throw InvalidPositionException();
	
	return block;
}

MapBlock * MapSector::createBlankBlockNoInsert(s16 y)
{
	// There should not be a block at this position
	if(getBlockBuffered(y) != NULL)
		throw AlreadyExistsException("Block already exists");

	v3s16 blockpos_map(m_pos.X, y, m_pos.Y);
	
	MapBlock *block = new MapBlock(m_parent, blockpos_map);
	
	return block;
}

MapBlock * MapSector::createBlankBlock(s16 y)
{
	JMutexAutoLock lock(m_mutex);
	
	MapBlock *block = createBlankBlockNoInsert(y);
	
	m_blocks.insert(y, block);

	return block;
}

void MapSector::insertBlock(MapBlock *block)
{
	s16 block_y = block->getPos().Y;

	{
		JMutexAutoLock lock(m_mutex);

		MapBlock *block2 = getBlockBuffered(block_y);
		if(block2 != NULL){
			throw AlreadyExistsException("Block already exists");
		}

		v2s16 p2d(block->getPos().X, block->getPos().Z);
		// FIXME: The program sometimes bumps here
		// NOTE: Haven't seen that lately.
		assert(p2d == m_pos);
		
		/*if(block_y > m_highest_y)
			m_highest_y = block_y;*/
	
		// Insert to container
		m_blocks.insert(block_y, block);
	}
}

core::list<MapBlock*> MapSector::getBlocks()
{
	JMutexAutoLock lock(m_mutex);

	core::list<MapBlock*> ref_list;

	core::map<s16, MapBlock*>::Iterator bi;

	bi = m_blocks.getIterator();
	for(; bi.atEnd() == false; bi++)
	{
		MapBlock *b = bi.getNode()->getValue();
		ref_list.push_back(b);
	}
	
	return ref_list;
}

f32 MapSector::getGroundHeight(v2s16 p, bool generate)
{
	if(m_heightmap == NULL)
		return GROUNDHEIGHT_NOTFOUND_SETVALUE;
	return m_heightmap->getGroundHeight(p);
}

void MapSector::setGroundHeight(v2s16 p, f32 y, bool generate)
{
	if(m_heightmap == NULL)
		return;
	m_heightmap->setGroundHeight(p, y);
	differs_from_disk = true;
}

void MapSector::serialize(std::ostream &os, u8 version)
{
	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: MapSector format not supported");
	
	/*
		[0] u8 serialization version
		+ heightmap data
	*/
	
	// Server has both of these, no need to support not having them.
	assert(m_heightmap != NULL);
	assert(m_objects != NULL);

	/*
		Save heightmap
	*/
	{
		// Write version
		os.write((char*)&version, 1);
		
		/*
			Write heightmap
		*/
		u32 heightmap_size =
				FixedHeightmap::serializedLength(version, MAP_BLOCKSIZE);
		
		SharedBuffer<u8> data(heightmap_size);
		m_heightmap->serialize(*data, version);

		os.write((const char*)*data, heightmap_size);
		
		if(version >= 5)
		{
			/*
				Write objects
			*/
			
			u16 object_count;
			if(m_objects->size() > 65535)
				object_count = 65535;
			else
				object_count = m_objects->size();

			u8 b[2];
			writeU16(b, object_count);
			os.write((char*)b, 2);
			
			core::map<v3s16, u8>::Iterator i;
			i = m_objects->getIterator();
			for(; i.atEnd() == false; i++)
			{
				v3s16 p = i.getNode()->getKey();
				u8 d = i.getNode()->getValue();
				u8 b[7];
				writeV3S16(&b[0], p);
				b[6] = d;
				os.write((char*)b, 7);
			}
		}
	}
}

MapSector* MapSector::deSerialize(
		std::istream &is,
		NodeContainer *parent,
		v2s16 p2d,
		Heightmap *master_hm,
		core::map<v2s16, MapSector*> & sectors
	)
{
	/*
		[0] u8 serialization version
		+ heightmap data
	*/

	/*
		Read stuff
	*/
	
	// Read version
	u8 version = SER_FMT_VER_INVALID;
	is.read((char*)&version, 1);
	
	if(!ser_ver_supported(version))
		throw VersionMismatchException("ERROR: MapSector format not supported");
	
	/*
		Read heightmap
	*/

	u32 heightmap_size =
			FixedHeightmap::serializedLength(version, MAP_BLOCKSIZE);
	
	SharedBuffer<u8> data(heightmap_size);
	is.read((char*)*data, heightmap_size);

	FixedHeightmap *hm = new FixedHeightmap(master_hm, p2d, MAP_BLOCKSIZE);
	hm->deSerialize(*data, version);

	/*
		Read objects
	*/

	core::map<v3s16, u8> *objects = new core::map<v3s16, u8>;

	if(version >= 5)
	{
		u8 b[2];
		is.read((char*)b, 2);
		u16 object_count = readU16(b);

		for(u16 i=0; i<object_count; i++)
		{
			u8 b[7];
			is.read((char*)b, 7);
			v3s16 p = readV3S16(&b[0]);
			u8 d = b[6];
			objects->insert(p, d);
		}
	}
	
	/*
		Get or create sector
	*/

	MapSector *sector = NULL;

	core::map<v2s16, MapSector*>::Node *n = sectors.find(p2d);

	if(n != NULL)
	{
		std::cerr<<"deSerializing existent sectors not supported "
				"at the moment, because code hasn't been tested."
				<<std::endl;
		assert(0);
		
		sector = n->getValue();
	}
	else
	{
		sector = new MapSector(parent, p2d);
		sectors.insert(p2d, sector);
	}

	/*
		Set stuff in sector
	*/
	
	// Set (or change) heightmap
	FixedHeightmap *oldhm = sector->m_heightmap;
	sector->m_heightmap = hm;
	if(oldhm != NULL)
		delete oldhm;
	
	// Set (or change) objects
	core::map<v3s16, u8> *oldfo = sector->m_objects;
	sector->m_objects = objects;
	if(oldfo)
		delete oldfo;

	return sector;
}

//END
