/*
(c) 2010 Perttu Ahola <celeron55@gmail.com>
*/

#ifndef MAPSECTOR_HEADER
#define MAPSECTOR_HEADER

#include <jmutex.h>
#include "common_irrlicht.h"
#include "mapblock.h"
#include "heightmap.h"
#include "exceptions.h"

/*
	This is an Y-wise stack of MapBlocks.
*/

#define WATER_LEVEL (-5)

#define HEIGHTMAP_OBJECT_TEST 0
#define HEIGHTMAP_OBJECT_TREE_1 1

class MapSector :
		public NodeContainer,
		public Heightmappish
{
private:
	
	// The pile of MapBlocks
	core::map<s16, MapBlock*> m_blocks;
	//JMutex m_blocks_mutex; // For public access functions

	NodeContainer *m_parent;
	// Position on parent (in MapBlock widths)
	v2s16 m_pos;

	// Be sure to set this to NULL when the cached block is deleted 
	MapBlock *m_block_cache;
	s16 m_block_cache_y;

	JMutex m_mutex;
	
	/*
		Stuff to be used when generating the map.
	*/
	FixedHeightmap *m_heightmap;
	// These are removed when they are drawn to blocks.
	// - Each is drawn when generating blocks; When the last one of
	//   the needed blocks is being generated.
	core::map<v3s16, u8> *m_objects;
	
	// Needed for lighting calculation (sunlight comes from top of this)
	//s16 m_highest_y;

	/*
		Private methods
	*/
	MapBlock *getBlockBuffered(s16 y);

public:
	
	// When true, sector metadata is changed from the one on disk
	// (sector metadata = all but blocks)
	// Basically, this should be changed to true in every setter method
	bool differs_from_disk;

	// Used by ServerMap. Counts seconds from last usage.
	// Sector can be deleted from memory after some time of inactivity.
	// NOTE: It has to be made very sure no other thread is accessing
	//       the sector and it doesn't remain in any cache when
	//       deleting it.
	float usage_timer;

	MapSector(NodeContainer *parent, v2s16 pos);
	virtual ~MapSector();

	/*FixedHeightmap * getHeightmap()
	{
		//TODO: See if setting this causes problems
		differs_from_disk = true;
		return m_heightmap;
	}*/

	void setHeightmap(FixedHeightmap *fh)
	{
		m_heightmap = fh;
		differs_from_disk = true;
	}

	void setObjects(core::map<v3s16, u8> *objects)
	{
		m_objects = objects;
		differs_from_disk = true;
	}

	core::map<v3s16, u8> * getObjects()
	{
		differs_from_disk = true;
		return m_objects;
	}

	v2s16 getPos()
	{
		return m_pos;
	}

	MapBlock * getBlockNoCreate(s16 y);
	MapBlock * createBlankBlockNoInsert(s16 y);
	MapBlock * createBlankBlock(s16 y);
	//MapBlock * getBlock(s16 y, bool generate=true);

	void insertBlock(MapBlock *block);
	
	core::list<MapBlock*> getBlocks();
	
	// These have to exist
	f32 getGroundHeight(v2s16 p, bool generate=false);
	void setGroundHeight(v2s16 p, f32 y, bool generate=false);
	
	// virtual from NodeContainer
	bool isValidPosition(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);

		if(blockpos.X != 0 || blockpos.Z != 0)
			return false;

		MapBlock *blockref;
		try{
			blockref = getBlockNoCreate(blockpos.Y);
		}
		catch(InvalidPositionException &e)
		{
			return false;
		}

		return true;
	}

	// virtual from NodeContainer
	MapNode getNode(v3s16 p)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		if(blockpos.X != 0 || blockpos.Z != 0)
			throw InvalidPositionException
				("MapSector only allows Y");

		MapBlock * blockref = getBlockNoCreate(blockpos.Y);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;

		return blockref->getNode(relpos);
	}
	// virtual from NodeContainer
	void setNode(v3s16 p, MapNode & n)
	{
		v3s16 blockpos = getNodeBlockPos(p);
		if(blockpos.X != 0 || blockpos.Z != 0)
			throw InvalidPositionException
				("MapSector only allows Y");

		MapBlock * blockref = getBlockNoCreate(blockpos.Y);
		v3s16 relpos = p - blockpos*MAP_BLOCKSIZE;
		blockref->setNode(relpos, n);
	}
	
	/*
		These functions handle metadata.
		They do not handle blocks.
	*/
	void serialize(std::ostream &os, u8 version);
	static MapSector* deSerialize(
			std::istream &is,
			NodeContainer *parent,
			v2s16 p2d,
			Heightmap *master_hm,
			core::map<v2s16, MapSector*> & sectors
		);
};

#endif

