#ifndef SERVER_HEADER
#define SERVER_HEADER

#include "connection.h"
#include "environment.h"
#include "common_irrlicht.h"
#include <string>

#ifdef _WIN32
	#include <windows.h>
	#define sleep_ms(x) Sleep(x)
#else
	#include <unistd.h>
	#define sleep_ms(x) usleep(x*1000)
#endif

struct QueuedBlockEmerge
{
	v3s16 pos;
	// key = peer_id, value = flags
	core::map<u16, u8> peer_ids;
};

/*
	This is a thread-safe class.
*/
class BlockEmergeQueue
{
public:
	BlockEmergeQueue()
	{
		m_mutex.Init();
	}

	~BlockEmergeQueue()
	{
		JMutexAutoLock lock(m_mutex);

		core::list<QueuedBlockEmerge*>::Iterator i;
		for(i=m_queue.begin(); i!=m_queue.end(); i++)
		{
			QueuedBlockEmerge *q = *i;
			delete q;
		}
	}

	void addBlock(u16 peer_id, v3s16 pos, u8 flags)
	{
		JMutexAutoLock lock(m_mutex);

		/*
			Find if block is already in queue.
			If it is, update the peer to it and quit.
		*/
		core::list<QueuedBlockEmerge*>::Iterator i;
		for(i=m_queue.begin(); i!=m_queue.end(); i++)
		{
			QueuedBlockEmerge *q = *i;
			if(q->pos == pos)
			{
				q->peer_ids[peer_id] = flags;
			}
		}
		
		/*
			Add the block
		*/
		QueuedBlockEmerge *q = new QueuedBlockEmerge;
		q->pos = pos;
		q->peer_ids[peer_id] = flags;
		m_queue.push_back(q);
	}

	// Returned pointer must be deleted
	// Returns NULL if queue is empty
	QueuedBlockEmerge * pop()
	{
		JMutexAutoLock lock(m_mutex);

		core::list<QueuedBlockEmerge*>::Iterator i = m_queue.begin();
		if(i == m_queue.end())
			return NULL;
		QueuedBlockEmerge *q = *i;
		m_queue.erase(i);
		return q;
	}

private:
	core::list<QueuedBlockEmerge*> m_queue;
	JMutex m_mutex;
};

class SimpleThread : public JThread
{
	bool run;
	JMutex run_mutex;

public:

	SimpleThread():
		JThread(),
		run(true)
	{
		run_mutex.Init();
	}

	virtual ~SimpleThread()
	{}

	virtual void * Thread() = 0;

	bool getRun()
	{
		JMutexAutoLock lock(run_mutex);
		return run;
	}
	void setRun(bool a_run)
	{
		JMutexAutoLock lock(run_mutex);
		run = a_run;
	}

	void stop()
	{
		setRun(false);
		while(IsRunning())
			sleep_ms(100);
	}
};

class Server;

class ServerThread : public SimpleThread
{
	Server *m_server;

public:

	ServerThread(Server *server):
		SimpleThread(),
		m_server(server)
	{
	}

	void * Thread();
};

class EmergeThread : public SimpleThread
{
	Server *m_server;

public:

	EmergeThread(Server *server):
		SimpleThread(),
		m_server(server)
	{
	}

	void * Thread();

	//TODO: I guess this can sometimes go wrong
	void trigger()
	{
		setRun(true);
		if(IsRunning() == false)
		{
			Start();
		}
	}
};

struct PlayerInfo
{
	u16 id;
	char name[PLAYERNAME_SIZE];
	v3f position;
	Address address;
	float avg_rtt;

	PlayerInfo();
	void PrintLine(std::ostream *s);
};

u32 PIChecksum(core::list<PlayerInfo> &l);

//TODO: Use these. Make a map of them with peer_ids as keys
struct RemoteClient
{
	RemoteClient()
	{
		peer_id = 0;
		serialization_version = SER_FMT_VER_INVALID;
	}
	~RemoteClient()
	{
	}
	
	// peer_id=0 means this client has no associated peer
	u8 peer_id;
	// The serialization version to use with the client
	u8 serialization_version;
};

class Server : public con::PeerHandler
{
public:
	/*
		NOTE: Every public method should be thread-safe
	*/
	Server(std::string mapsavedir);
	~Server();
	void start(unsigned short port);
	void stop();
	void step(float dtime);
	void AsyncRunStep();
	void Receive();
	void ProcessData(u8 *data, u32 datasize, u16 peer_id);
	core::list<PlayerInfo> getPlayerInfo();

private:

	// Virtual methods from con::PeerHandler.
	// As of now, these create and remove clients and players.
	// TODO: Make it possible to leave players on server.
	void peerAdded(con::Peer *peer);
	void deletingPeer(con::Peer *peer, bool timeout);
	
	// Helper methods
	void SendPlayerPositions();
	void SendPlayerInfos();
	RemoteClient* getClient(u16 peer_id);
	
	Environment m_env;
	JMutex m_env_mutex;

	con::Connection m_con;
	JMutex m_con_mutex;
	
	core::map<u16, RemoteClient*> m_clients;
	JMutex m_clients_mutex;

	float m_step_dtime;
	JMutex m_step_dtime_mutex;

	ServerThread m_thread;
	EmergeThread m_emergethread;

	BlockEmergeQueue m_emerge_queue;

	friend class EmergeThread;
};

#endif

